from typing import List

STATS_HEADER: List[str] = [
    "sample",
    "Flowcell",
    "Lanes",
    "readcounts/lane",
    "sum_readcounts",
    "yieldMB/lane",
    "sum_yield",
    "%Q30",
    "MeanQscore",
]
